#include "align.h"
#include <algorithm>
#include <vector>
#include <deque>
#include <cmath>
#include "image_types.h"
#include "constants.h"
#include "fit.h"
#include "transform.h"
/*
static const uint8_t star_begin_threshold(30);
static const uint8_t star_end_threshold(15);
static const size_t  star_min_size(10);
static const size_t  star_max_width(20);
static const size_t  star_max_size(star_max_width*star_max_width);

static const double one_eighty_over_pi(57.2957795131); // degrees to radians
*/

#define PROJECT_DOWN
bool s3d_merge::operator<(const s3d_merge::StarRef &a, const s3d_merge::StarRef &b)
{
    return (a.GetBrightness() > b.GetBrightness());
}

void _CollectAndZeroRegion(s3d_merge::Image8 *img, 
                           uint32_t seed_x, uint32_t seed_y, 
                           double star_end_threshold, bool use_weight, 
                           double &mean_x, double &mean_y, 
                           double &total_weight, size_t &pixels)
{
    std::deque<std::pair<uint32_t,uint32_t> > queue;

    queue.push_front(std::make_pair(seed_x, seed_y));

    while(queue.size())
    {
        uint32_t x, y;
        x = queue.back().first;
        y = queue.back().second;
        queue.pop_back();

        if((x < img->get_width()) && (y < img->get_height()))
        {
            if(img->get_pixel(x, y)[0] > star_end_threshold)
            {
                double weight(1.0);
    
                if(use_weight)
                {
                    weight = img->get_pixel(x, y)[0];
                }
    
                pixels++;
                mean_x+= double(x)*weight;
                mean_y+= double(y)*weight;
                total_weight+= weight;
    
                {
                    // Zero pixel.
                    uint8_t val(0);
                    img->set_pixel(x, y, &val);
                }
    
                if(x && (img->get_pixel(x-1, y)[0] > star_end_threshold))
                {
                    queue.push_front(std::make_pair(x-1, y));
                }
                if(y && (img->get_pixel(x, y-1)[0] > star_end_threshold))
                {
                    queue.push_front(std::make_pair(x, y-1));
                }
                if((x < (img->get_width() - 1)) && (img->get_pixel(x+1, y)[0] > star_end_threshold))
                {
                    queue.push_front(std::make_pair(x+1, y));
                }
                if((y < (img->get_height() - 1)) && (img->get_pixel(x, y+1)[0] > star_end_threshold))
                {
                    queue.push_front(std::make_pair(x, y+1));
                }
            }
        }
    }
}

void
CollectAndZeroRegion(s3d_merge::Image8 *img, 
                     uint32_t x, uint32_t y, double star_end_threshold, 
                     double &mean_x, double &mean_y, 
                     double &total_weight, size_t &pixels)
{
    pixels = 0;
    mean_x = 0.0;
    mean_y = 0.0;
    total_weight = 0.0;
    
    if(img)
    {
        _CollectAndZeroRegion(img, x, y, star_end_threshold, true, mean_x, mean_y, total_weight, pixels);
        mean_x/=total_weight;
        mean_y/=total_weight;
    }
}

void BlurHorizontal(const s3d_merge::Image8 *in, s3d_merge::Image8 *out, uint32_t radius)
{
    const uint32_t width(in->get_width());
    const uint32_t total_size((radius * 2) + 1);
    const uint32_t right(width - (radius + 1));
    
    for(int32_t y=0;y<(int32_t)in->get_height();y++)
    {
        uint32_t total;
        uint8_t *pixel;
        const uint8_t *front, *back;
        int32_t x;
    
        // first set up the line.
        pixel = out->get_buffer() + (y * width);
        back = in->get_buffer() + (y * width);
        front = back + radius;
    
        total = 0.0;
    
        for(x=(0 - int32_t(radius));x<=int32_t(radius);x++)
        {
            total+= *(in->get_constrained_pixel(x, y, false, false));
        }
    
        if(total_size < width)
        {
            for(x=0;x<int32_t(radius);x++)
            {
                *pixel++ = total / total_size;
                total-= *(in->get_constrained_pixel(x-radius, y, false, false));
                total+= *++front;
            }
    
            for(x = radius;x<int32_t(right);x++)
            {
                *pixel++ = total / total_size;
                total-= *back++;
                total+= *++front;
            }
    
            for(x = right;x<int32_t(width);x++)
            {
                *pixel++ = total / total_size;
                total-= *back++;
                total+= *(in->get_constrained_pixel(x+radius+1, y, false, false));
            }
        }
        else
        {
            for(x=0;x<int32_t(in->get_width());x++)
            {
                *pixel++ = total / total_size;
                total-= *(in->get_constrained_pixel(x-radius, y, false, false));
                total+= *(in->get_constrained_pixel(x+radius+1, y, false, false));
            }
        }
    }
}

void BlurVertical(const s3d_merge::Image8 *in, s3d_merge::Image8 *out, uint32_t radius)
{
    const uint32_t width(in->get_width());
    //const uint32_t height(in->get_height());
    const uint32_t total_size((radius * 2) + 1);
    //const uint32_t bottom(height - (radius + 1));

    for(int32_t x=0;x<int32_t(in->get_width());x++)
    {
        uint32_t total;
        uint8_t *pixel;
        /*const uint8_t *front, *back;*/
        int32_t y;

        // first set up the line.
        pixel = out->get_buffer() + x;
        //back = in->get_buffer() + x;
        //front = back + (radius * width);

        total = 0.0;

        for(y=(0 - (int32_t)radius);y<=(int32_t)radius;y++)
        {
            total+= *(in->get_constrained_pixel(x, y, false, false));
        }

        for(y=0;y<int32_t(in->get_height());y++)
        {
            *pixel = total/total_size;
            pixel+= width;
            // move forward.
            total-= *(in->get_constrained_pixel(x, y-radius, false, false));
            total+= *(in->get_constrained_pixel(x, y+radius+1, false, false));
        }
    }
}

s3d_merge::Image8 *BlurImage(s3d_merge::Image8 *img, uint32_t radius, size_t passes = 1)
{
    s3d_merge::Image8 *in = new s3d_merge::Image8(*img);
    s3d_merge::Image8 *out = new s3d_merge::Image8(*img);

    for(size_t i=0;i<passes;i++)
    {
        BlurHorizontal(in, out, radius);
        std::swap(in, out);
    }

    for(size_t i=0;i<passes;i++)
    {
        BlurVertical(in, out, radius);
        std::swap(in, out);
    }
    std::swap(in, out);

    delete in;
    return out;
}

// Locate stars in an image.
void 
s3d_merge::FindStars(const Image16 *img, 
                     size_t max_stars, 
                     uint8_t star_begin_threshold, 
                     uint8_t star_end_threshold, 
                     size_t star_min_size,
                     size_t star_max_size,
                     std::vector<StarRef> &stars)

{
    if(img && img->get_width() && img->get_height())
    {
        s3d_merge::Image8 *bw_img = new s3d_merge::Image8(img->get_width(), img->get_height(), 1);
        if(bw_img)
        {
            size_t size(img->get_width() * img->get_height());
            
            {
                const uint16_t *in(img->get_buffer());
                uint8_t *out(bw_img->get_buffer());
                uint32_t mean;
    
                for(size_t i=0;i<size;i++)
                {
                    mean = 0;
                    mean+=(*in++);
                    mean+=(*in++);
                    mean+=(*in++);
                    mean/=3;
                    mean>>= 8;
                    *out++ = (mean & 0xff);
                }
            }
    
            {
                s3d_merge::Image8 *filtered = BlurImage(bw_img, 100);
    
                const uint8_t *in = filtered->get_buffer();
                uint8_t *out = bw_img->get_buffer();
    
                for(size_t i=0;i<size;i++)
                {
                    if(*in > *out)
                    {
                        *out = 0;
                    }
                    else
                    {
                        *out-= *in;
                    }
                    in++;
                    out++;
                }
                delete filtered;
            }
    
            {
                uint8_t *p = bw_img->get_buffer();
    
                for(uint32_t y=0;y<bw_img->get_height();y++)
                {
                    for(uint32_t x=0;x<bw_img->get_width();x++)
                    {
                        if(*p > star_begin_threshold)
                        {
                            double mean_x, mean_y;
                            size_t pixels;
                            double total;
                            CollectAndZeroRegion(bw_img, x, y, star_end_threshold, mean_x, mean_y, total, pixels);
    
                            if((pixels > star_min_size) &&
                               (pixels < star_max_size))
                            {
                                stars.push_back(StarRef(mean_x, mean_y, total/255, pixels));
                            }
                        }
                        p++;
                    }
                }
            }
            delete bw_img;      
        }
        if(stars.size())
        {
            std::sort(stars.begin(), stars.end());
        }
        if(max_stars)
        {
            if(stars.size() > max_stars)
            {
                stars.erase(stars.begin()+max_stars, stars.end());
            }
        }
    }
}

// Functions to generate pattern nodes in 2 dimensions.
s3d_merge::PatternNode
s3d_merge::GeneratePatternNode2D(const std::vector<s3d_merge::Point2D> &points, const s3d_merge::KDTree2D *tree, size_t index)
{
    PatternNode result;

    if(tree && points.size())
    {

        std::vector<size_t> found;
        
        found.clear();
        tree->FindClosestItems(points[index], 5, found);

        if(found.size() == 5)
        {
            Point2D vectors[4];
            
            for(size_t j=0;j<4;j++)
            {
                vectors[j] = points[found[j+1]] - points[index];
                vectors[j].Normalise();
            }
            // Permutations.  4! = 12, but dp is commutative, so 6d output.
            result[0] = Point2D::DotProduct(vectors[0], vectors[1]);
            result[1] = Point2D::DotProduct(vectors[0], vectors[2]);
            result[2] = Point2D::DotProduct(vectors[0], vectors[3]);
            result[3] = Point2D::DotProduct(vectors[1], vectors[2]);
            result[4] = Point2D::DotProduct(vectors[1], vectors[3]);
            result[5] = Point2D::DotProduct(vectors[2], vectors[3]);

            for(size_t j=0;j<6;j++)
            {
                result[j] = acos(result[j]);
            }
        }
    }
    return result;
}

void
s3d_merge::GeneratePatternNodes2D(const std::vector<s3d_merge::Point2D> &points,
                                  const s3d_merge::KDTree2D *tree,
                                  std::vector<s3d_merge::PatternNode> &out)
{
    if(tree && points.size())
    {
        out.reserve(points.size());

        for(size_t i=0;i<points.size();i++)
        {
            out.push_back(GeneratePatternNode2D(points, tree, i));
        }
    }
}

// Functions to generate pattern nodes in 3 dimensions.  3D vectors are points on unit sphere.
s3d_merge::PatternNode
s3d_merge::GeneratePatternNode3D(const std::vector<s3d_merge::Point3D> &points,
                                 const s3d_merge::KDTree3D *tree,
                                 size_t index)
{
    PatternNode result;

    if(tree && points.size())
    {

        std::vector<size_t> found;
        
        found.clear();
        tree->FindClosestItems(points[index], 5, found);

        if(found.size() == 5)
        {
            Point3D vectors[4];
            
            for(size_t j=0;j<4;j++)
            {
                vectors[j] = points[found[j+1]] - points[index];
                vectors[j].Normalise();
                // Project onto plane normal to star direction.
#ifdef PROJECT_DOWN
                vectors[j]-= (points[index] * Point3D::DotProduct(points[index], vectors[j]));
                vectors[j].Normalise();
#endif
            }
            // Permutations.  4! = 12, but dp is commutative, so 6d output.
            result[0] = Point3D::DotProduct(vectors[0], vectors[1]);
            result[1] = Point3D::DotProduct(vectors[0], vectors[2]);
            result[2] = Point3D::DotProduct(vectors[0], vectors[3]);
            result[3] = Point3D::DotProduct(vectors[1], vectors[2]);
            result[4] = Point3D::DotProduct(vectors[1], vectors[3]);
            result[5] = Point3D::DotProduct(vectors[2], vectors[3]);

            for(size_t j=0;j<6;j++)
            {
                result[j] = acos(result[j]);
            }
        }
    }
    return result;
}

void
s3d_merge::GeneratePatternNodes3D(const std::vector<s3d_merge::Point3D> &points,
                                  const s3d_merge::KDTree3D *tree,
                                  std::vector<s3d_merge::PatternNode> &out)
{
    if(tree && points.size())
    {
        out.reserve(points.size());

        for(size_t i=0;i<points.size();i++)
        {
            out.push_back(GeneratePatternNode3D(points, tree, i));
        }
    }
}


// Constrained 2D pattern generation
bool
s3d_merge::GeneratePatternNode2D(const std::vector<s3d_merge::Point2D> &points,
                                 const s3d_merge::KDTree2D *tree,
                                 size_t index,
                                 double min_radius,
                                 double max_radius,
                                 s3d_merge::PatternNode &result)
{
    bool created(false);

    min_radius = fabs(min_radius);    
    max_radius = fabs(max_radius);    

    if(min_radius > max_radius)
    {
        std::swap(min_radius, max_radius);
    }

    if(tree && points.size())
    {
        std::vector<size_t> found;
        tree->FindItemsInRange(points[index], max_radius, found);

        if(found.size() > 4)
        {
            std::vector<size_t> filtered;

            for(size_t j=0;j<found.size();j++)
            {
                if((found[j] != index) && 
                   (Point2D::Distance(points[found[j]], points[index]) > min_radius))
                {
                    filtered.push_back(found[j]);
                }
            }

            if(filtered.size() >= 4)
            {
                // Get 4 brightest points.
                Point2D vectors[4];
                std::sort(filtered.begin(), filtered.end());
            
                for(size_t j=0;j<4;j++)
                {
                    vectors[j] = points[filtered[j]] - points[index];
                    vectors[j].Normalise();
                }

                // Permutations.  4! = 12, but dp is commutative, so 6d output.
                result[0] = Point2D::DotProduct(vectors[0], vectors[1]);
                result[1] = Point2D::DotProduct(vectors[0], vectors[2]);
                result[2] = Point2D::DotProduct(vectors[0], vectors[3]);
                result[3] = Point2D::DotProduct(vectors[1], vectors[2]);
                result[4] = Point2D::DotProduct(vectors[1], vectors[3]);
                result[5] = Point2D::DotProduct(vectors[2], vectors[3]);
    
                for(size_t j=0;j<6;j++)
                {
                    result[j] = acos(result[j]);
                }
                created = true;
            }
        }
    }

    return created;
}

void
s3d_merge::GeneratePatternNodes2D(const std::vector<s3d_merge::Point2D> &points,
                                  const s3d_merge::KDTree2D *tree,
                                  double min_radius,
                                  double max_radius,
                                  std::vector<s3d_merge::PatternNode> &out,
                                  std::vector<size_t> &mapping)
{
    for(size_t i=0;i<points.size();i++)
    {
        PatternNode tmp;

        if(GeneratePatternNode2D(points, tree, i, min_radius, max_radius, tmp))
        {
            out.push_back(tmp);
            mapping.push_back(i);
        }
    }
}

// Constrained 3D pattern generation
bool
s3d_merge::GeneratePatternNode3D(const std::vector<s3d_merge::Point3D> &points,
                      const s3d_merge::KDTree3D *tree,
                      size_t index,
                      double min_radius,
                      double max_radius,
                      s3d_merge::PatternNode &result)
{
    bool created(false);

    min_radius = fabs(min_radius);    
    max_radius = fabs(max_radius);    

    if(min_radius > max_radius)
    {
        std::swap(min_radius, max_radius);
    }

    if(tree && points.size())
    {
        std::vector<size_t> found;
        tree->FindItemsInRange(points[index], max_radius, found);

        if(found.size() > 4)
        {
            std::vector<size_t> filtered;
            for(size_t j=0;j<found.size();j++)
            {
                if((found[j] != index) && 
                   (Point3D::Distance(points[found[j]], points[index]) > min_radius))
                {
                    filtered.push_back(found[j]);
                }
            }
            if(filtered.size() >= 4)
            {
                // Get 4 brightest points.
                Point3D vectors[4];
                std::sort(filtered.begin(), filtered.end());
            
                for(size_t j=0;j<4;j++)
                {
                vectors[j] = points[found[j+1]] - points[index];
                vectors[j].Normalise();
                // Project onto plane normal to star direction.
#ifdef PROJECT_DOWN
                vectors[j]-= (points[index] * Point3D::DotProduct(points[index], vectors[j]));
                vectors[j].Normalise();
#endif
                }

                // Permutations.  4! = 12, but dp is commutative, so 6d output.
                result[0] = Point3D::DotProduct(vectors[0], vectors[1]);
                result[1] = Point3D::DotProduct(vectors[0], vectors[2]);
                result[2] = Point3D::DotProduct(vectors[0], vectors[3]);
                result[3] = Point3D::DotProduct(vectors[1], vectors[2]);
                result[4] = Point3D::DotProduct(vectors[1], vectors[3]);
                result[5] = Point3D::DotProduct(vectors[2], vectors[3]);
    
                for(size_t j=0;j<6;j++)
                {
                    result[j] = acos(result[j]);
                }
                created = true;
            }
        }
    }

    return created;
}

void
s3d_merge::GeneratePatternNodes3D(const std::vector<s3d_merge::Point3D> &points,
                       const s3d_merge::KDTree3D *tree,
                       double min_radius,
                       double max_radius,
                       std::vector<s3d_merge::PatternNode> &out,
                       std::vector<size_t> &mapping)
{
    for(size_t i=0;i<points.size();i++)
    {
        PatternNode tmp;

        if(GeneratePatternNode3D(points, tree, i, min_radius, max_radius, tmp))
        {
            out.push_back(tmp);
            mapping.push_back(i);
        }
    }
}

class RansacData
{
public:

    RansacData(const std::vector<std::pair<s3d_merge::Point2D, s3d_merge::Point2D> > &matches,
               s3d_merge::Transform &transform)
        : m_matches(matches)
        , m_transform(transform)
    {
    }

    const std::vector<std::pair<s3d_merge::Point2D, s3d_merge::Point2D> > &m_matches;
    s3d_merge::Transform &m_transform;
};

static double RansacError(const void *p, const size_t *items, size_t nitems, const double *parameters, size_t /*nparameters*/)
{
    double result(0);

    if(p && parameters)
    {
        RansacData *data = (RansacData*)p;

        data->m_transform.SetParameters(parameters);

        for(size_t i=0;i<nitems;i++)
        {
            s3d_merge::Point2D tmp;
            
            data->m_transform.FwdTransformPoint(data->m_matches[items[i]].first, tmp);

            tmp-= data->m_matches[items[i]].second;

            result+= s3d_merge::Point2D::DotProduct(tmp, tmp);
        }
        result/= nitems;
    }

    return result;
}

bool 
s3d_merge::FitStars(const s3d_merge::PatternTree *ref_pattern_tree,
                    const std::vector<s3d_merge::PatternNode> &/*ref_patterns For further analysis.  TODO*/,
                    const s3d_merge::KDTree2D *ref_tree,
                    const std::vector<s3d_merge::Point2D> &ref_points,
                    const std::vector<s3d_merge::Point2D> &cur_points,
                    double consensus_error,
                    s3d_merge::Transform &transform)
{
    bool matched(false);

    double *alignment;

    alignment = new double[transform.GetNumParameters()]; 

    if(ref_pattern_tree && ref_tree && ref_points.size() && cur_points.size() && alignment)
    {
        size_t ransac_count = transform.GetNumParameters() + 3;

        transform.GetParameters(alignment);
        for(size_t i=0;i<transform.GetNumParameters();i++)
        {
            printf("%g, ", alignment[i]);
        }
        putchar('\n');

        std::vector<s3d_merge::PatternNode> cur_patterns;

        {
            KDTree2D cur_tree(cur_points.begin(), cur_points.end());
            GeneratePatternNodes2D(cur_points, &cur_tree, cur_patterns);
        }

        if(cur_patterns.size())
        {
            std::vector<std::pair<Point2D, Point2D> > matches;

            for(size_t i=0;i<cur_patterns.size();i++)
            {
                size_t found;
                double distance;
                
                if(ref_pattern_tree->FindClosestItem(cur_patterns[i], found, distance))
                {
                    if(distance < 0.1)
                    {
                        double dist;

                        dist = Point2D::DotProduct((cur_points[i] - ref_points[found]),
                                                   (cur_points[i] - ref_points[found]));
                        dist = sqrt(dist);

                        matches.push_back(std::make_pair(cur_points[i], ref_points[found]));
                    }
                }
            }


            if(matches.size() > (ransac_count * 2)) 
            {
                size_t target_consensus = (matches.size() / 2);
                size_t acceptable_consensus = transform.GetNumParameters() * 2;

                printf("Attempting to fit transform to %li matches\n", matches.size());
                RansacData data(matches, transform);
                size_t fits = fit_ransac(
                    RansacError, (void *)&data, matches.size(), alignment, transform.GetNumParameters(),
                    0.01, 1.0e-6, 2000, // Least square fit
                    ransac_count, (consensus_error*consensus_error), target_consensus, 5000);
 
                if(fits >= acceptable_consensus) // TODO: Magic number elimination.
                {
                    printf("Fitted %li points\n", fits);
                    matched = true;
                }
                else
                {
                    printf("Unable to fit image\n");
                }
            }
            else
            {
                printf("Insufficient matches for RANSAC\n");
            }
        }
        transform.GetParameters(alignment);
        for(size_t i=0;i<transform.GetNumParameters();i++)
        {
            printf("%g, ", alignment[i]);
        }
        putchar('\n');
    }

    if(alignment)
    {
        delete[] alignment;
    }

    return matched;
}

