#ifndef __CATALOG_H__
#define __CATALOG_H__

#include <string>
#include <stdint.h>
#include <vector>
#include <functional>

namespace s3d_merge
{ 

class BSCLine
{
public:
    BSCLine(const char *p)
        : m_p(p)
    {
    }

    const char *m_p;
};

class CatalogueRef
{
public:
    CatalogueRef()
        : m_valid(false)
        , m_bsc_id(-1)
        , m_decj2000(0)
        , m_raj2000(0)
        , m_magnitude(0)
    {
    }

    CatalogueRef(const BSCLine &line);

    CatalogueRef(const CatalogueRef &other)
        : m_valid(other.GetValid())
        , m_bsc_id(other.GetBSCID())
        , m_decj2000(other.GetDecJ2000())
        , m_raj2000(other.GetRaJ2000())
        , m_magnitude(other.GetRawMagnitude())
        , m_name(other.GetName())
    {
    }

    const CatalogueRef &operator=(const CatalogueRef &other)
    {
        m_valid      = other.GetValid();
        m_bsc_id     = other.GetBSCID();
        m_decj2000   = other.GetDecJ2000();
        m_raj2000    = other.GetRaJ2000();
        m_magnitude  = other.GetRawMagnitude();
        m_name       = other.GetName();
        return *this;
    }

    // Processed content
    double GetMagnitude(void) const; 
    double GetRelativeIntensity(void) const; 
    void Get3DVector(double *) const;

    void GetRightAscention(std::string &) const;
    void GetDeclination(std::string &) const;

    // Exact content
    bool GetValid(void) const
    {
        return m_valid;
    }
    uint32_t GetBSCID(void) const
    {
        return m_bsc_id;
    }

    int32_t GetDecJ2000(void) const
    {
        return m_decj2000;
    }

    int32_t GetRaJ2000(void) const
    {
        return m_raj2000;
    }

    int32_t GetRawMagnitude(void) const
    {
        return m_magnitude;
    }

    const std::string &GetName(void) const
    {
        return m_name;
    }

private:
    bool     m_valid;
    uint32_t m_bsc_id;
    int32_t m_decj2000; // arc seconds
    int32_t m_raj2000;  // tenths of seconds
    int32_t m_magnitude;
    std::string m_name;

    static bool GetField(const char *, size_t, size_t, size_t, char *);
};

class MagnitudeComparator : public std::binary_function<const CatalogueRef &, const CatalogueRef &, bool>
{
public:
    bool operator()(const CatalogueRef &a, const CatalogueRef &b)
    {
        return (a.GetRawMagnitude() < b.GetRawMagnitude());
    }
};


void ParseYaleCatalogue(const char *fn, std::vector<CatalogueRef> &out);


}

#endif

