#ifndef __IMAGE_STACK_H__
#define __IMAGE_STACK_H__

#include "image.h"
#include "image_types.h"
#include <stdint.h>
#include <string>
#include <vector>

#define MAX_IMAGES 32
namespace s3d_merge
{

class ImageData
{
public:
    enum PixelCategory
    {
        PIXEL_IN_RANGE,
        PIXEL_DARK,
        PIXEL_SATURATED,
        PIXEL_INVALID
    };

    // Classify a pixel in the input.
    static PixelCategory ClassifyValue(uint16_t val);


    ImageData(const char *fn, Image16 *image, double ev);
    ~ImageData();

    // Get the Exposure Value for this frame
    double EV(void) const;

    // Get the multiplier associated with the exposure value for this frame.
    double Multiplier(void) const;

    // Get the raw 16 bit image data for this frame.
    const Image16 *Image(void) const;

    // Classify a pixel in the input.
    PixelCategory ClassifyPixel(uint32_t x, uint32_t y, uint32_t channel) const;

    // Get weight of a pixel, given a set of weighting flags.
    double GetWeight(uint32_t x, uint32_t y, uint32_t channel, uint32_t flags) const;

    // Get inverted value of a pixel.
//    double GetValue(uint32_t x, uint32_t y, uint32_t channel,
//                    const CameraResponse &response) const;

    double GetDoubleValue(uint32_t x, uint32_t y, uint32_t channel) const;
    uint16_t GetValue(uint32_t x, uint32_t y, uint32_t channel) const;

private:
    std::string m_name;
    Image16    *m_image;
    double      m_ev;
    double      m_multiplier;
};

}
#endif


