#include "kdtree.h"
#include <iostream>
#include <vector>

// Test code for kd tree

using namespace spectral;

typedef Point<float, 2> MyPoint;

int main(int , char **)
{
    std::vector<MyPoint> points;
    std::vector<size_t> found, found2;

    MyPoint foo;

    for(size_t i=0;i<500000;i++)
    {
        foo[0] = rand();
        foo[0]/= RAND_MAX;
        foo[1] = rand();
        foo[1]/= RAND_MAX;

        points.push_back(foo);
    }

    std::cerr << "building tree" << std::endl;
    KDTree<float, 2> tree(points.begin(), points.end());

    std::cerr << "built tree" << std::endl;
    MyPoint min, max, centre;
    min[0] = 0.48;
    min[1] = 0.48;
    max[0] = 0.52;
    max[1] = 0.52;

    centre[0] = centre[1] = 0.5;

    for(size_t i=0;i<1000;i++)
    {
        found.clear();
        tree.FindItemsInRange(min, max, found);
    }
    std::cout << "Found " << found.size() << " of " << points.size() << " points" << std::endl;

    for(size_t i=0;i<1000;i++)
    {
        found2.clear();
        tree.FindItemsInRangeBF(min, max, found2);
    }
    std::cout << "Tree BF Found " << found2.size() << " of " << points.size() << " points" << std::endl;

    if(found.size() == found2.size())
    {
        std::sort(found.begin(), found.end());
        std::sort(found2.begin(), found2.end());
        for(size_t i=0;i<found.size();i++)
        {
            if(found[i] != found2[i])
            {
                std::cerr << "Results differ!" << std::endl;
            }
        }
    }
   
    size_t closest;
    size_t closest2;
    for(size_t i=0;i<1000;i++)
    {
        float distance;
        tree.FindClosestItem(centre, closest, distance);
    }
    std::cerr << "closest item = " << closest << " (" 
              << points[closest][0] << ", "
              << points[closest][1] << ")" << std::endl;

    for(size_t i=0;i<1000;i++)
    {
        float distance;
        tree.FindClosestItemBF(centre, closest2, distance);
    }

    std::cerr << "BF closest item = " << closest2 << " (" 
              << points[closest2][0] << ", "
              << points[closest2][1] << ")" << std::endl;

    std::vector<size_t> items;

    tree.FindClosestItems(centre, 10, items);
    for(size_t i=0;i<items.size();i++)
    {
    std::cerr << i << " closest item = " << items[i] << " (" 
              << points[items[i]][0] << ", "
              << points[items[i]][1] << ")" << std::endl;
    }
    return 0;
}
