#include "align.h"
#include "image_data.h"
#include "image_io.h"
#include "preferences.h"
#include "version.h"
#include <stdio.h>
#include <unistd.h>
#include <FreeImage.h>

void FreeImageErrorHandler(FREE_IMAGE_FORMAT fif,
                           const char *message)
{

    if(fif != FIF_UNKNOWN)
    {
        fprintf(stderr, "%s Format\n", FreeImage_GetFormatFromFIF(fif));
    }
    fprintf(stderr, "%s\n", message);
}


void show_help(void)
{
    printf("Star List %s. Copyright (C) David Beynon 2014\n\n", SAS_VERSION);
    printf("Usage: sas_list [options] <input>\n");
    printf("Options:\n");
    printf("\t-h        : This help text.\n");
    printf("\t-p <file> : Load preferences file.\n\n");
    printf("Output format: X, Y, Brightness, Size (pixels)\n\n");
    printf("Output is sorted brightest first.\n\n");
    printf("\n%s\n", FreeImage_GetCopyrightMessage());
}


int main(int argc, char ** argv)
{
    s3d_merge::Preferences prefs;

    if(argc == 1)
    {
        show_help();
    }
    else
    {
        {
            int c;
            while((c = getopt(argc, argv, "h,p:")) != EOF)
            {
                switch(c)
                {
                case 'h':
                    show_help();
                    exit(0);
                    break;
                case 'p':
                    prefs.LoadPreferences(optarg);
                    break;
                }
            }
        }
        if(optind < argc)
        {
            s3d_merge::ImageData *id;
    
            id = s3d_merge::LoadImageData(argv[optind]);
    
            if(id && id->Image())
            {
                std::vector<s3d_merge::StarRef> stars;
    
                FindStars(id->Image(),
                          0,
                          prefs.GetFindStarsTriggerThreshold(),
                          prefs.GetFindStarsExitThreshold(),
                          prefs.GetFindStarsMinPixels(),
                          prefs.GetFindStarsMaxPixels(),
                          stars);
    
                s3d_merge::FindStars(id->Image(), 0, 30, 15, 10, 400, stars);
    
                for(size_t i=0;i<stars.size();i++)
                {
                    printf("%g, %g, %g, %li\n", 
                           stars[i].GetPoint()[0], stars[i].GetPoint()[1], 
                           stars[i].GetBrightness(),
                           stars[i].GetPixelCount());
                }
            }
        }
    }
    return 0;
}
