#ifndef __MATCH_H__
#define __MATCH_H__

#include <stdint.h>
#include <vector>
#include <algorithm>
#include "kdtree.h"
#include "align.h"

namespace s3d_merge
{

class Triangle
{
public:
    Triangle()
    {
        m_points[0] = m_points[1] = m_points[2] = 0;
    }
    Triangle(size_t p1, size_t p2, size_t p3)
    {
        m_points[0] = p1;
        m_points[1] = p2;
        m_points[2] = p3;
    }
    Triangle(const Triangle &other)
    {
        *this = other;
    }
    const Triangle &operator=(const Triangle &other)
    {
        m_points[0] = other.m_points[0];
        m_points[1] = other.m_points[1];
        m_points[2] = other.m_points[2];
        return *this;
    }

    size_t m_points[3];
};

static bool operator<(const Triangle &a, const Triangle &b)
{
    // Inefficient
    if(a.m_points[0] >= b.m_points[0])
        return false;
    if(a.m_points[1] >= b.m_points[1])
        return false;
    if(a.m_points[2] >= b.m_points[2])
        return false;

    return true;
}

static bool operator==(const Triangle &a, const Triangle &b)
{
    return((a.m_points[0] == b.m_points[0]) &&
           (a.m_points[1] == b.m_points[1]) &&
           (a.m_points[2] == b.m_points[2]));
}


double CalculateAngleThreshold(double angle);

void BuildTriangles3D(const s3d_merge::KDTree3D &tree, 
                      const std::vector<s3d_merge::Point3D> &points,
                      double min_angle,
                      double max_angle,
                      size_t triangles_per_point, 
                      std::vector<Triangle> &triangles);

void BuildTriangles2D(const s3d_merge::KDTree2D &tree, 
                      const std::vector<s3d_merge::Point2D> &points,
                      double min_angle,
                      double max_angle,
                      size_t triangles_per_point, 
                      std::vector<Triangle> &triangles);

void
MatchTriangleShapes(const Triangle &t3d,
                    const std::vector<s3d_merge::Point3D> p3d,
                    const Triangle &t2d,
                    const std::vector<s3d_merge::Point2D> p2d,
                    Triangle &out_3d, 
                    Triangle &out_2d);
s3d_merge::Point2D
GetTriangleShape2D(const s3d_merge::Point2D &p1,
                   const s3d_merge::Point2D &p2,
                   const s3d_merge::Point2D &p3);


s3d_merge::Point2D
GetTriangleShape3D(const s3d_merge::Point3D &p1, 
                   const s3d_merge::Point3D &p2,
                   const s3d_merge::Point3D &p3);

void
BuildTriangleStats2D(const std::vector<Triangle> &triangles, 
                     const std::vector<s3d_merge::Point2D> &points,
                     std::vector<s3d_merge::Point2D> &shapes,
                     std::vector<s3d_merge::Point2D> &centroids);

void
BuildTriangleStats3D(const std::vector<Triangle> &triangles, 
                     const std::vector<s3d_merge::Point3D> &points,
                     std::vector<s3d_merge::Point2D> &shapes,
                     std::vector<s3d_merge::Point3D> &centroids);

};
#endif
