#ifndef __MERGE_UTILS_H__
#define __MERGE_UTILS_H__
#include "image_types.h"
#include <cassert>
namespace s3d_merge
{
    class Transform;

    class Merge
    {
    public:
        Merge();
        Merge(const FloatImage *);
        virtual ~Merge();

        virtual bool AddFrame(const Image16 *) = 0;  
        virtual bool PostProcess(void) = 0;

        FloatImage *GetImage(void) const
        {
            return m_image;
        }

    protected:
        bool SetImage(const Image16 *);

        const FloatImage *GetDark(void) const
        {
            return m_dark;
        }

    private:
        FloatImage *m_image;
        const FloatImage *m_dark;
    };


    class Average : public Merge
    {
    public:
        Average();
        Average(const FloatImage *);
        virtual ~Average();

        virtual bool AddFrame(const Image16 *);  
        virtual bool PostProcess(void);
    private:
        float m_count;
    };

    class Lighten : public Merge
    {
    public:
        Lighten();
        Lighten(const FloatImage *);
        virtual ~Lighten();

        virtual bool AddFrame(const Image16 *);  
        virtual bool PostProcess(void);
    private:
    };


    // TODO: Sort out hierachy
    class AverageAlign 
    {
    public:
        AverageAlign(const FloatImage *dark, const Image16 *first)
            : m_dark(dark)
            , m_image(NULL)
            , m_weights(NULL)
        {
            if(first)
            {
                m_weights = new FloatImage(first->get_width(), first->get_height(), 1);
                {
                    size_t size(m_weights->get_width() * m_weights->get_height());
                    for(size_t i=0;i<size;i++)
                    {
                        m_weights->get_buffer()[i] = 1.0;
                    }
                }
                m_image = MakeFloatImage(first);
            }
            else
            {
                assert(0);
            }
        }

        ~AverageAlign()
        {
            if(m_weights)
            {
                delete m_weights;
            }
            if(m_image)
            {
                delete m_image;
            }
        }
    
        
        bool AddFrame(const Image16 *img, uint32_t aa, const Transform &);
        void PostProcess(void);
    
        FloatImage *GetImage(void) const
        {
            return m_image;
        }
    
        private:
            FloatImage *MakeFloatImage(const Image16 *);
            const FloatImage *m_dark;
            FloatImage *m_image;
            FloatImage *m_weights;
    };
};


#endif
