#include "preferences.h"
#include <cstdio>
#include <cctype>
#include <cstring>
#include <cmath>

s3d_merge::Preferences::Preferences()
{
    SetDefaults();
}

s3d_merge::Preferences::Preferences(const char *fn)
{
    SetDefaults();

    LoadPreferences(fn);
}

void 
s3d_merge::Preferences::SetDefaults(void)
{
    m_findStarsTriggerThreshold = 30;
    m_findStarsExitThreshold    = 15;
    m_findStarsMinPixels        = 10;
    m_findStarsMaxPixels        = 400;
    m_findStarsMaxStars         = 200;
    m_findStarsBlurRadius       = 100;

    m_alignStarsConsensusError  = 3;

    m_remapImageAAType          = 1;
}

void
s3d_merge::Preferences::LoadPreferences(const char *fn)
{
    if(fn)
    {
        FILE *fp = fopen(fn, "r");
        if(fp)
        {

            while(!feof(fp))
            {
                char tmp[1024];
    
                fgets(tmp, 1024, fp);
    
                if(tmp[0])
                {
                    char *sep;
                    char *p, *q;
                    p = q = tmp;
    
                    do
                    {
                        if(isspace(*p))
                        {
                            p++;
                        }
                        else
                        {
                            *q++ = tolower(*p++);
                        }
                    } while(*p);
    
                    sep = strchr(tmp, '=');
    
                    if(sep)
                    {
                        *sep = '\0';
                        sep++;
    
                        if(!ParseField(tmp, sep))
                        {
                            printf("Failed to parse line \"%s\", \"%s\n", tmp, sep);
                        }
                    }
                }
            }
            fclose(fp);
        }
        else
        {
            fprintf(stderr, "Unable to open file \"%s\"\n", fn);
        }
    }
}

static const char *s_findStarsTriggerThreshold("find_stars_trigger_threshold");
static const char *s_findStarsExitThreshold("find_stars_exit_threshold");
static const char *s_findStarsMinPixels("find_stars_min_pixels");
static const char *s_findStarsMaxPixels("find_stars_max_pixels");
static const char *s_findStarsMaxStars("find_stars_max_stars");
static const char *s_findStarsBlurRadius("find_stars_blur_radius");

static const char *s_alignStarsConsensusError("align_stars_consensus_error");

static const char *s_remapImageAAType("remap_image_aa_type");

bool
s3d_merge::Preferences::ParseField(const char *field, const char *value)
{
    bool parsed(false);

    if(field && *field && value && *value)
    {
        if(!strcmp(field, s_findStarsTriggerThreshold))
        {
            uint32_t tmp(atoi(value));
            if(tmp > 255) tmp = 255;

            m_findStarsTriggerThreshold = uint8_t(tmp);
            parsed = true;
        }
        else if(!strcmp(field, s_findStarsExitThreshold))
        {
            uint32_t tmp(atoi(value));
            if(tmp > 255) tmp = 255;

            m_findStarsExitThreshold = uint8_t(tmp);
            parsed = true;
        }
        else if(!strcmp(field, s_findStarsMinPixels))
        {
            m_findStarsMinPixels = atoi(value);
            parsed = true;
        }
        else if(!strcmp(field, s_findStarsMaxPixels))
        {
            m_findStarsMaxPixels = atoi(value);
            parsed = true;
        }
        else if(!strcmp(field, s_findStarsMaxStars))
        {
            m_findStarsMaxStars = atoi(value);
            parsed = true;
        }
        else if(!strcmp(field, s_findStarsBlurRadius))
        {
            m_findStarsBlurRadius = atoi(value);
            if(!m_findStarsBlurRadius)
            {
                m_findStarsBlurRadius = 1;
            }
            parsed = true;
        }
        else if(!strcmp(field, s_alignStarsConsensusError))
        {
            m_alignStarsConsensusError = fabs(atof(value));
            parsed = true;
        }
        else if(!strcmp(field, s_remapImageAAType))
        {
            m_remapImageAAType = atoi(value);
            if(m_remapImageAAType > 1) // BLEAH!
                m_remapImageAAType = 1;
            parsed = true;
        }
    }

    return parsed;
}

void
s3d_merge::Preferences::DumpPreferences(void)
{
    printf("%s=%i\n",
           s_findStarsTriggerThreshold, m_findStarsTriggerThreshold);

     printf("%s=%i\n",
            s_findStarsExitThreshold, m_findStarsExitThreshold);

     printf("%s=%li\n",
            s_findStarsMinPixels,
            m_findStarsMinPixels);

     printf("%s=%li\n",
            s_findStarsMaxPixels,
            m_findStarsMaxPixels);

     printf("%s=%li\n",
            s_findStarsMaxStars,
            m_findStarsMaxStars);
     
     printf("%s=%i\n",
            s_findStarsBlurRadius,
            m_findStarsBlurRadius);

     printf("%s=%g\n",
            s_alignStarsConsensusError,
            m_alignStarsConsensusError);

     printf("%s=%i\n",
            s_remapImageAAType,
            m_remapImageAAType);
}
