#ifndef __TRANSFORM_H__
#define __TRANSFORM_H__

#include "align.h"

namespace s3d_merge{

// Transform class designed to be used with fitter.

class Transform
{
public:
    Transform();
    virtual ~Transform();

    void SetParameters(const double *);

    virtual void GetParameters(double *) const;

    virtual void OnSetParameters()                           = 0;
    virtual void SetDefaults()                               = 0;
    virtual size_t GetNumParameters() const                  = 0;
    
    virtual void FwdTransformPoint(const Point2D &, Point2D &) const = 0;
    virtual void InvTransformPoint(const Point2D &, Point2D &) const = 0;
protected:
    double *m_data;
};

class RotateAndTranslate : public Transform
{
    enum
    {
        RAT_TRANSLATE_X = 0,
        RAT_TRANSLATE_Y,
        RAT_ANGLE,
        RAT_MAX
    };
public:
    RotateAndTranslate();
    RotateAndTranslate(const RotateAndTranslate &other);

    virtual ~RotateAndTranslate();

    const RotateAndTranslate &operator=(const RotateAndTranslate &other);

    virtual void OnSetParameters();
    virtual void SetDefaults();
    virtual size_t GetNumParameters() const;
    
    virtual void FwdTransformPoint(const Point2D &, Point2D &) const;
    virtual void InvTransformPoint(const Point2D &, Point2D &) const;
private:
    Point2D m_fwd_i, m_fwd_j, m_t, m_inv_i, m_inv_j;
};

class LensDistort : public Transform
{
    enum
    {
        //LD_X = 0,
        //LD_Y,
        LD_K1 = 0,
        LD_K2,
        LD_K3,
        LD_K4,
        LD_MAX
       /* LD_P1,
        LD_P2,
        LD_P3,
        LD_P4,*/
    };
public:

    LensDistort();
    LensDistort(uint32_t x_res, uint32_t y_res);
    LensDistort(const LensDistort &other);

    virtual ~LensDistort();

    const LensDistort &operator=(const LensDistort &other);

    virtual void OnSetParameters();
    virtual void SetDefaults();
    virtual size_t GetNumParameters() const;
    
    virtual void FwdTransformPoint(const Point2D &, Point2D &) const;
    virtual void InvTransformPoint(const Point2D &, Point2D &) const;

private:
    double FwdRadialTerm(double) const;
    double InvRadialTerm(double) const;

    Point2D m_centre;
    Point2D m_offset;
    double m_radius;
};

class TransformAndLens : public Transform
{
public:
    TransformAndLens(uint32_t x_res, uint32_t y_res);
    virtual ~TransformAndLens();

    virtual void GetParameters(double *) const;
    virtual void OnSetParameters();
    virtual void SetDefaults();
    virtual size_t GetNumParameters() const;
    
    virtual void FwdTransformPoint(const Point2D &, Point2D &) const;
    virtual void InvTransformPoint(const Point2D &, Point2D &) const;

    Transform &GetMove(void) const
    {
        return *(Transform*)(&m_move);
    }
private:
    LensDistort        m_lens;
    RotateAndTranslate m_move;
    
};

};


#endif
